(this.webpackJsonp=this.webpackJsonp||[]).push([[10],{1121:function(e,t,a){"use strict";a.r(t);a(0);var n=a(983),c=a(227),s=a(335),l=a(139),o=a(391),i=a(5);t.default=function(){var e=Object(o.g)({sortBy:l.c.LastExecutionTime,perPage:1}),t=e.loading,a=e.tasks,u=Object(o.f)(),r=u.loading,b=u.tasks,d=t||r;return Object(i.e)(n.a,null,Object(i.e)(s.a,{loading:d},Object(i.e)(c.a,null,b.length," in Queue"),!!b.length&&Object(i.e)(c.a,null,Object(i.e)("b",null,"Currently Executing"),": ",b[0].name),!!a.length&&Object(i.e)(c.a,null,Object(i.e)("b",null,"Last Task Execution"),": ",a[0].name," (",new Date(a[0].lastExecutionTime).toLocaleString(),")")))}},983:function(e,t,a){"use strict";var n=a(3),c=a(13),s=a(0),l=a.n(s),o=(a(24),a(15)),i=a(19),u=l.a.forwardRef((function(e,t){var a=e.classes,s=e.className,i=e.component,u=void 0===i?"div":i,r=Object(c.a)(e,["classes","className","component"]);return l.a.createElement(u,Object(n.a)({className:Object(o.a)(a.root,s),ref:t},r))}));t.a=Object(i.a)({root:{padding:16,"&:last-child":{paddingBottom:24}}},{name:"MuiCardContent"})(u)}}]);
//# sourceMappingURL=TasksHomeCard.0f1a9b4992c80e636a4f.js.map